import org.ejml.simple.SimpleMatrix;

import java.util.ArrayList;
import java.util.function.BiFunction;

public class Network {

    ArrayList<Layer> layers;
    BiFunction<SimpleMatrix, SimpleMatrix, Double> loss;
    BiFunction<SimpleMatrix, SimpleMatrix, SimpleMatrix> lossPrime;

    public Network() {
        layers = new ArrayList<>();
    }

    public void add(Layer layer) {
        layers.add(layer);
    }

    public void use(BiFunction<SimpleMatrix, SimpleMatrix, Double> loss, BiFunction<SimpleMatrix, SimpleMatrix, SimpleMatrix> lossPrime) {
        this.loss = loss;
        this.lossPrime = lossPrime;
    }

    public SimpleMatrix[] predict(SimpleMatrix[] inputs) {
        SimpleMatrix[] result = new SimpleMatrix[inputs.length];
        SimpleMatrix output;
        int i = 0;

        for (SimpleMatrix input : inputs) {
            output = input;
            for (Layer l : layers) {
                output = l.forwardPropagation(output);
            }
            result[i] = output;
            i++;
        }

        return result;
    }

    public void fit(SimpleMatrix[] X_train, SimpleMatrix[] y_train, int epochs, double learningRate) {
        int samples = X_train.length;

        for (int i = 0; i < epochs; i++) {
            double err = 0;
            for (int j = 0; j < samples; j++) {
                // forward propagation
                SimpleMatrix output = X_train[j];
                for (Layer l : layers) {
                    output = l.forwardPropagation(output);
                }

                // compute loss (for display purpose only)
                err = loss.apply(y_train[j], output);

                // backward propagation
                SimpleMatrix error = lossPrime.apply(y_train[j], output);
                for (int k = layers.size() - 1; k >= 0; k--) {
                    error = layers.get(k).backwardPropagation(error, learningRate);
                }
            }
            // calculate average error on all samples
            err /= samples;
            System.out.println("epoch " + (i+1) + "/" + epochs + "  error=" + err);
        }
    }
}